﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.Ports;
using System.Diagnostics;
using System.Reflection;
using System.Threading;

namespace AVRBootLoaderNamespace
   {
   class AVRBootLoader
      {
      private enum eBootloaderCommands
         {
         REVISION = 0,	// get bootloader revision
         BUFFSIZE = 1,	// get buffer size
         SIGNATURE = 2,	// get target signature
         USERFLASH = 3,	// get user flash size
         PROGRAM = 4,	// program flash
         START = 5,	   // start application
         CHECK_CRC = 6,	// CRC o.k.
         VERIFY = 7	   // Verify
         }

      // Protokoll Definitionen
      #region "Protokoll"
      private const System.Byte COMMAND = 0xA5;		// Command sequence start
      private const System.Byte ESCAPE = COMMAND;

      private const System.Byte CONNECT = 0xA6;	   // connection established
      private const System.Byte BADCOMMAND = 0xA7;	// command not supported
      private const System.Byte ANSWER = 0xA8;	      // followed by length byte
      private const System.Byte CONTINUE = 0xA9;
      private const System.Byte SUCCESS = 0xAA;
      private const System.Byte FAIL = 0xAB;

      private const System.Byte ESC_SHIFT = 0x80;		// offset escape char
      private const System.Byte PROGEND = ESC_SHIFT;
      #endregion

      // Diverse Konstanten
      private const int TIMEOUT = 300;		// 0.3s
      private const int TIMEOUTP = 4000;	// 4s
      private const int STARTWAIT = 200;

      private const System.UInt32 MAXFLASH = 0x40000;	// max flash size (256kB)

      private int Echocount = 0;
      private int _Buffersize = 0;
      private int _Flashsize = 0;
      private System.UInt16 Crc = 0;

      private int _DeviceID = 0;
      public int DeviceID
         {
         get { return this._DeviceID; }
         }

      private int _BootloaderRevision = 0;
      public int BootloaderRevision
         {
         get { return this._BootloaderRevision; }
         }

      public int FlashSize
      {
          get { return this._Flashsize; }
      }

      public int Buffersize
      {
          get { return this._Buffersize; }
      }

      private string _Password = "";
      //public string _Password = 
      public string Password
         {
          get { return this._Password; }
          set { this._Password = value; }
         }

      private bool _OneWire = false;
      public bool OneWire
         {
         get { return this._OneWire; }
         }

      private string _ErrorMessage = "";
      public string ErrorMessage
         {
         get { return this._ErrorMessage; }
         set { this._ErrorMessage = value; }
         }

      private bool _Busy = false;
      public bool Busy
         {
         get { return this._Busy; }
         }

      private bool _Success = false;
      public bool Success
         {
         get { return this._Success; }
         }

      private int _BytesWritten = 0;
      public int BytesWritten
         {
         get { return this._BytesWritten; }
         }

      private bool _Aborted = false;
      public bool Aborted
         {
         get { return this._Aborted; }
         }

      private bool AbortThread;

      // Abbild des Flash-Speichers
      private System.Byte[] Data;

      // Serieller Port
      private SerialPort MySerialPort;

      private bool Connected = false;
      private bool CrcSupported = false;

      private string sFilename = "";
      private bool bVerify=false;

      public delegate void ProgressChangedEventHandler(int ProgressPercent);
      public event ProgressChangedEventHandler ProgressChanged;

      public AVRBootLoader(SerialPort MySerialPort)
         {
            this.ProgressChanged = null;

            if (MySerialPort == null)
                throw new ArgumentException("MySerialPort nicht gesetzt");

            this.MySerialPort = MySerialPort;
            this.Data = new System.Byte[AVRBootLoader.MAXFLASH];
         }

      public void AVRBootloaderAbort()
         {
         this.AbortThread = true;
         }

      public void AVRBootloaderConnect()
         {
         ThreadStart ts = new ThreadStart(this._AVRBootloaderConnect);

         Thread t = new Thread(ts);

         this.AbortThread = false;
         this._Busy = true;
         this._Success = false;

         t.Start();
         }

      private void _AVRBootloaderConnect()
         {
         int crc_on;

         this.Crc = 0;
         this._BytesWritten = 0;

         if(this.SerialConnect() != 0)	// connect ATMega
            {
            this.ErrorMessage = "Connect abgebrochen";
            this._Busy = false;
            return;
            }

         crc_on = this.CheckCrc();	// 2 = no CRC support
         if(crc_on != 2)
            this.CrcSupported = true;
         // CRC nochmals synchronisieren (wg. Problemen beim Connect)
         crc_on = this.CheckCrc();	// 2 = no CRC support

         if(this.ReadInfo() != 0)
            {
            this._Busy = false;
            return;
            }

         if(this.CrcSupported && this.CheckCrc() != 0)
            {
            this.ErrorMessage = "Fehler beim ersten Überprüfen des CRC";
            this._Busy = false;
            return;
            }

         if(this.AbortThread)
            {
            this._Aborted = true;
            this._ErrorMessage = "Vorgang abgebrochen";
            this._Busy = false;
            return;
            }

         this.Connected = true;

         this._Success = true;
         this._Busy = false;

         return;
         }

      public string AvrGetDeviceNameFromID(string sdevicename, Int32 ideviceid)
      {
          switch (ideviceid)
          {
              case 0x1e9307: sdevicename = "ATMega 8"; break;
              case 0x1e9403: sdevicename = "ATmega16"; break;
              case 0x1e940c: sdevicename = "ATmega16HVA"; break;
              case 0x1e9502: sdevicename = "ATmega32"; break;
              case 0x1e9510: sdevicename = "ATmega32HVB"; break;
              case 0x1e9205: sdevicename = "ATmega48"; break;
              case 0x1e920a: sdevicename = "ATmega48P"; break;
              case 0x1e9602: sdevicename = "ATmega64"; break;
              case 0x1e930a: sdevicename = "ATmega88"; break;
              case 0x1e930f: sdevicename = "ATmega88P"; break;
              case 0x1e9701: sdevicename = "ATmega103"; break;
              case 0x1e9702: sdevicename = "ATmega128"; break;
              case 0x1e9401: sdevicename = "ATmega161"; break;
              case 0x1e9404: sdevicename = "ATmega162"; break;
              case 0x1e9402: sdevicename = "ATmega163"; break;
              case 0x1e940a: sdevicename = "ATmega164P"; break;              
              case 0x1e9407: sdevicename = "ATmega165(P)"; break;
              case 0x1e9406: sdevicename = "ATmega168"; break;
              case 0x1e940b: sdevicename = "ATmega168P"; break;
              case 0x1e9405: sdevicename = "ATmega169(P)"; break;
              case 0x1e9501: sdevicename = "ATmega323"; break;
              case 0x1e9508: sdevicename = "ATmega324P"; break;
              case 0x1e9505: sdevicename = "ATmega325"; break;
              case 0x1e955d: sdevicename = "ATmega325P"; break;
              case 0x1e950f: sdevicename = "ATmega328P"; break;
              case 0x1e9503: sdevicename = "ATmega329(P)"; break;
              case 0x1e9507: sdevicename = "ATmega406"; break;
              case 0x1e9608: sdevicename = "ATmega640"; break;
              case 0x1e9609: sdevicename = "ATmega644"; break;
              case 0x1e960a: sdevicename = "ATmega644P"; break;
              case 0x1e9605: sdevicename = "ATmega645"; break;
              case 0x1e9603: sdevicename = "ATmega649"; break;
              case 0x1e9703: sdevicename = "ATmega1280"; break;
              case 0x1e9704: sdevicename = "ATmega1281"; break;
              case 0x1e9705: sdevicename = "ATmega1284P"; break;
              case 0x1e9801: sdevicename = "ATmega2560"; break;
              case 0x1e9802: sdevicename = "ATmega2561"; break;
              case 0x1e9506: sdevicename = "ATmega3250"; break;
              case 0x1e956e: sdevicename = "ATmega3250P"; break;
              case 0x1e950c: sdevicename = "ATmega3290"; break;
              case 0x1e9504: sdevicename = "ATmega3290P"; break;
              case 0x1e9606: sdevicename = "ATmega6450"; break;
              case 0x1e9604: sdevicename = "ATmega6490"; break;
              case 0x1e9306: sdevicename = "ATmega8515"; break;
              case 0x1e9308: sdevicename = "ATmega8535"; break;
              case 0x1e9383: sdevicename = "AT90PWM2B"; break;
              case 0x1e9381: sdevicename = "AT90PWMx(B)"; break;
              case 0x1e9483: sdevicename = "AT90PWMx16"; break;              
              case 0x1e9584: sdevicename = "AT90PWM324"; break;
              case 0x1e9004: sdevicename = "ATtiny11"; break;
              case 0x1e9005: sdevicename = "ATtiny12"; break;
              case 0x1e9007: sdevicename = "ATtiny13"; break;
              case 0x1e9006: sdevicename = "ATtiny15"; break;
              case 0x1e9106: sdevicename = "ATtiny22"; break;
              case 0x1e910b: sdevicename = "ATtiny24"; break;
              case 0x1e9108: sdevicename = "ATtiny25"; break;
              case 0x1e9109: sdevicename = "ATtiny26"; break;
              case 0x1e9107: sdevicename = "ATtiny28"; break;
              case 0x1e920c: sdevicename = "ATtiny43U"; break;
              case 0x1e9207: sdevicename = "ATtiny44"; break;
              case 0x1e9206: sdevicename = "ATtiny45"; break;
              case 0x1e9209: sdevicename = "ATtiny48"; break;
              case 0x1e930c: sdevicename = "ATtiny84"; break;
              case 0x1e930b: sdevicename = "ATtiny85"; break;
              case 0x1e9311: sdevicename = "ATtiny88"; break;
              case 0x1e910c: sdevicename = "ATtiny261"; break;
              case 0x1e9208: sdevicename = "ATtiny461"; break;
              case 0x1e930d: sdevicename = "ATtiny861"; break;
              case 0x1e910a: sdevicename = "ATtiny2313"; break;
              case 0x1e9382: sdevicename = "AT90USB82"; break;              
              case 0x1e9482: sdevicename = "AT90USB162"; break;
              case 0x1e9682: sdevicename = "AT90USB646/7"; break;
              case 0x1e9782: sdevicename = "AT90USB1286/7"; break;              
              case 0x1e9001: sdevicename = "AT90S1200"; break;
              case 0x1e9101: sdevicename = "AT90S2313"; break;
              case 0x1e9102: sdevicename = "AT90S2323"; break;              
              case 0x1e9103: sdevicename = "AT90S2343"; break;
              case 0x1e9201: sdevicename = "AT90S4414"; break;
              case 0x1e9203: sdevicename = "AT90S4433"; break;
              case 0x1e9303: sdevicename = "AT90S4434/8535"; break;
              case 0x1e9301: sdevicename = "AT90S8515"; break;
              case 0x1e9181: sdevicename =  "AT86RF401"; break;
              case 0x1e9581: sdevicename =  "AT90CAN32"; break;
              case 0x1e9681: sdevicename =  "AT90CAN64"; break;
              case 0x1e9781: sdevicename =  "AT90CAN128"; break; 
              
              default: sdevicename = "devicename not found!"; break;
          }
          return sdevicename;
      }


      public void AvrBootloaderProgram(string sFilename)
         {
         ThreadStart ts = new ThreadStart(this._AvrBootloaderProgram);

         Thread t = new Thread(ts);

         this.sFilename = sFilename;
         this.bVerify = false;

         this.AbortThread = false;
         this._Busy = true;
         this._Success = false;

         t.Start();
         }

      private void _AvrBootloaderProgram()
         {
         if(!this.Connected)
            {
            this._Busy = false;
            return;
            }

         if(this.ProcessFile(this.sFilename, this.bVerify) != 0)
            {
            this._Busy = false;
            return;
            }

         if(this.CrcSupported && this.CheckCrc() != 0)
            {
            this.ErrorMessage = "CRC-Fehler";
            this._Busy = false;
            return;
            }

         this._Success = true;
         this._Busy = false;

         return;
         }

      public void AvrBootloaderVerify(string sFilename)
         {
         ThreadStart ts = new ThreadStart(this._AvrBootloaderProgram);

         Thread t = new Thread(ts);

         this.sFilename = sFilename;
         this.bVerify = true;

         this.AbortThread = false;
         this._Busy = true;
         this._Success = false;

         t.Start();
         }

      public void AVRBootloaderDisconnect()
         {
         ThreadStart ts = new ThreadStart(this._AVRBootloaderDisconnect);

         Thread t = new Thread(ts);

         this.AbortThread = false;
         this._Busy = true;
         this._Success = false;

         t.Start();
         }

      private void _AVRBootloaderDisconnect()
         {
         if(!this.Connected)
            {
            this._Busy = false;
            return;
            }

         this.SerialExit();

         this.Connected = false;
         this._Success = true;
         this._Busy = false;

         return;
         }

      private int SerialConnect()
         {
         int nReceiveChar = 0;
         int echo = 0;
         int nCurrentPasswordChar = 0;

         this._OneWire = false;
         for(; ; )
            {
            nCurrentPasswordChar = 0;
            do
               {
               if(nCurrentPasswordChar < this.Password.Length)
                  this.Send((System.Byte)this.Password[nCurrentPasswordChar]);
               else
                  this.Send(0xFF);
               nReceiveChar = this.Receive(0);
               if(nReceiveChar == this.Password[1])
                  echo = 1;		// echo received

               if(nReceiveChar == AVRBootLoader.CONNECT)
                  {
                  if(echo != 0)
                     {
                     this.Echocount = 1;
                     this._OneWire = true;
                     }
                  this.SendCommand(AVRBootLoader.COMMAND);
                  for(; ; )
                     {
                     switch(this.Receive(AVRBootLoader.TIMEOUT))
                        {		// clear RX buffer
                        case AVRBootLoader.SUCCESS:
                        case -1:
                           return 0;
                        }
                     }
                  }
               nCurrentPasswordChar++;
               if(this.AbortThread)
                  {
                  this._Aborted = true;
                  return 1;
                  }
               Thread.Sleep(0);
               }
            while(nCurrentPasswordChar <= this.Password.Length);
            }
         }

      private void SerialExit()
         {
         this.SendCommand(AVRBootLoader.PROGEND);
         this.SendCommand(AVRBootLoader.COMMAND);
         this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.START);
         }

      private int ReadInfo()
         {
         this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.REVISION);
         this._BootloaderRevision = this.ReadVal();
          Trace.TraceInformation(this.GetType().Name + "." + MethodInfo.GetCurrentMethod().Name.ToString() + "(): "
            + String.Format("Bootloader Revision V{0:X}.{1:X}", this._BootloaderRevision >> 8, this._BootloaderRevision & 0xff));

         this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.SIGNATURE);
         this._DeviceID = this.ReadVal();
         if(this._DeviceID == -1)
            return 1;
         Trace.TraceInformation(this.GetType().Name + "." + MethodInfo.GetCurrentMethod().Name.ToString() + "(): "
            + String.Format("DeviceID: {0:X}", this._DeviceID));

         this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.BUFFSIZE);
         this._Buffersize = this.ReadVal();
         if(this._Buffersize == -1)
            return 1;
         Trace.TraceInformation(this.GetType().Name + "." + MethodInfo.GetCurrentMethod().Name.ToString() + "(): "
            + String.Format("Buffersize: {0}", this._Buffersize));

         this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.USERFLASH);
         this._Flashsize = this.ReadVal();
         if(this._Flashsize == -1 || this._Flashsize > MAXFLASH)
            return 1;
         Trace.TraceInformation(this.GetType().Name + "." + MethodInfo.GetCurrentMethod().Name.ToString() + "(): "
            + String.Format("Flashsize: {0}", this._Flashsize));
         return 0;
         }

      private int ProcessFile(string fname, bool bVerify)
         {
         System.UInt32 lastaddr = 0;
         System.UInt32 addr = 0;
         int i;
         int nCurrentProgress, nLastProgress = 0;
         HexFile MyHexFile = new HexFile();

         System.Byte d1;

         this._ErrorMessage = "";

         if(!System.IO.File.Exists(fname))
            {
            this._ErrorMessage = string.Format("Die Datei '{0}' existiert nicht", fname);
            return 1;
            }

         try
            {
            MyHexFile.LoadFile(fname);
            }
         catch(Exception e)
            {
            this._ErrorMessage = string.Format("Fehler '{0}' beim Laden der Datei", e.Message);
            return 1;
            }

         for(i = 0; i < MyHexFile.HexFileLines.Count; i++)
            {

            if(MyHexFile.HexFileLines[i].Address + MyHexFile.HexFileLines[i].NumBytes > this._Flashsize)
               {
               this._ErrorMessage = "Hex-Datei zu groß für Device";
               return 1;
               }
            if(MyHexFile.HexFileLines[i].RecordType == HexFile.eRecordType.DataRecord)
               {
               Array.Copy(MyHexFile.HexFileLines[i].Data, 0,
                  this.Data, MyHexFile.HexFileLines[i].Address, MyHexFile.HexFileLines[i].NumBytes);
               addr = (System.UInt32)(MyHexFile.HexFileLines[i].Address + MyHexFile.HexFileLines[i].NumBytes);
               if(lastaddr < addr - 1)
                  lastaddr = addr - 1;
               addr++;
               }
            }
         if(!bVerify)
            {
            this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.PROGRAM);
            }
         else
            {
            this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.VERIFY);
            if(this.Receive(TIMEOUT) == AVRBootLoader.BADCOMMAND)
               {
               this._ErrorMessage = "Verify nicht verfügbar";
               return 1;
               }
            }
         nLastProgress = 0;
         if(this.ProgressChanged != null)
            this.ProgressChanged(0);

         byte[] tmpbuffer = new byte[_Buffersize*2];      // tem. Sendbuffer, size = max for data+ESC
         int n=0;

         for(i = this._Buffersize, addr = 0; ; addr++)
            {
            switch(d1 = this.Data[addr])
               {
               case ESCAPE:
                  goto case 0x13;
               case 0x13:
                  //this.Send(AVRBootLoader.ESCAPE);
                  tmpbuffer[n++] = AVRBootLoader.ESCAPE;
                  UpdateCrc(AVRBootLoader.ESCAPE);      // calculate transmit CRC
                  d1 += ESC_SHIFT;
                  goto default;
               default:
                  //this.Send(d1);
                  tmpbuffer[n++] = d1;
                  UpdateCrc(d1);      // calculate transmit CRC
                  break;
               }
            if(--i == 0)
               {
                if (n >0)
                {
                    MySerialPort.Write(tmpbuffer, 0, n);
                    n = 0;
                }

               if(!bVerify && this.Receive(TIMEOUTP) != AVRBootLoader.CONTINUE)
                  {
                  this._ErrorMessage = String.Format("Programmieren bei Adresse {0} fehlgeschlagen", addr);
                  return 1;
                  }
               i = this._Buffersize;
               }
            if(addr == lastaddr)
               {
                if (n > 0)
                {
                    MySerialPort.Write(tmpbuffer, 0, n);
                    n = 0;
                }

               this.Send(AVRBootLoader.ESCAPE);
               this.Send(AVRBootLoader.ESC_SHIFT);	// A5,80 = End
               if(this.Receive(TIMEOUTP) != AVRBootLoader.SUCCESS)
                  {
                  if(bVerify)
                     this._ErrorMessage = "Überprüfen fehlgeschlagen";
                  else
                     this._ErrorMessage = "Programmieren fehlgeschlagen";
                  return 1;
                  }
               break;
               }
            if(this.AbortThread)
               {
               this._Aborted = true;
               this._ErrorMessage = "Vorgang abgebrochen";
               return 1;
               }
            // ProgressEvent z.B. für Fortschrittsanzeige
            nCurrentProgress = (int)((float)addr / lastaddr * 100);
            if(nCurrentProgress != nLastProgress)
               {
               nLastProgress = nCurrentProgress;
               if(this.ProgressChanged != null)
                  this.ProgressChanged(nCurrentProgress);
               }
            Thread.Sleep(0);
            }
         if(nLastProgress != 100)
            {
            if(this.ProgressChanged != null)
               this.ProgressChanged(100);
            }
         this._BytesWritten = (int)lastaddr;

         return 0;
         }


      private byte[] acBuffer = new byte[1];
      private void Send(System.Byte vcByte)
         {
         if(this.Echocount != 0)
            {
            if(this.Echocount > 1)
               this.Receive(0);		// remove echo
            this.Echocount++;
            }
         acBuffer[0] = vcByte;
         this.MySerialPort.Write(acBuffer, 0, 1);
         this.UpdateCrc(vcByte);			// calculate transmit CRC
         }

      private System.Int16 Receive(int vnTimeout)
         {
         this.MySerialPort.ReadTimeout = vnTimeout;
         try
            {
            return (System.Int16)this.MySerialPort.ReadByte();
            }
         catch
            {
            return -1;
            }
         }

      private void SendCommand(System.Byte vcCommand)
         {
         if(this.Echocount != 0)
            this.Echocount = 1;		// restart echo counter
         this.Send(AVRBootLoader.COMMAND);
         this.Send(vcCommand);
         }

      private void UpdateCrc(System.Byte vcByte)
         {
         int i;

         this.Crc ^= vcByte;
         for(i = 8; i > 0; i--)
            {
            this.Crc = (System.UInt16)((this.Crc >> 1)
               ^ (((this.Crc & 1) != 0) ? 0xA001 : 0));
            }
         }

      private int CheckCrc()
         {
         System.UInt16 Crc;

         this.SendCommand((System.Byte)AVRBootLoader.eBootloaderCommands.CHECK_CRC);
         if(this.Receive(AVRBootLoader.TIMEOUT) == AVRBootLoader.BADCOMMAND)
            return 2;
         Crc = this.Crc;
         this.Send((System.Byte)(Crc & 0xff));
         this.Send((System.Byte)(Crc >> 8));
         if(this.Receive(AVRBootLoader.TIMEOUT) == AVRBootLoader.SUCCESS)
            return 0;
         return 1;
         }

      private System.Int32 ReadVal()
         {
         int i;
         int j = 257;
         System.Int32 val = 0;

         for(; ; )
            {
            i = this.Receive(AVRBootLoader.TIMEOUT);
            if(i == -1)
               return -1;		// timeout

            switch(j)
               {
               case 1:
                  if(i == AVRBootLoader.SUCCESS)
                     return val;
                  goto default;
               case 2:
               case 3:
               case 4:
                  val = val * 256 + i;
                  j--;
                  break;
               case 256:
                  j = i;
                  break;
               case 257:
                  if(i == AVRBootLoader.FAIL)
                     return -2;
                  if(i == AVRBootLoader.ANSWER)
                     j = 256;
                  break;
               default:
                  return -2;
               }
            }
         }

      }

   // Daten einer kompletten Hex-Datei
   internal class HexFile
      {
      public enum eRecordType
         {
         DataRecord = 0,
         EndOfFileRecord = 1,
         ExtendedSegmentAddressRecord = 2,
         StartSegmentAddressRecord = 3,
         ExtendedLinearAddressRecord = 4,
         StartLinearAddressRecord = 5
         }

      private string _Filename = "";
      public string Filename
         {
         get { return this._Filename; }
         }

      private List<HexFileLine> _HexFileLines;
      public List<HexFileLine> HexFileLines
         {
         get { return this._HexFileLines; }
         }

      // Konstruktor
      public HexFile()
         {
         this._HexFileLines = new List<HexFileLine>();
         }

      public void LoadFile(string vsFilename)
         {
         string sLine;
         FileStream fs;
         StreamReader sr;

         if(!System.IO.File.Exists(vsFilename))
            throw new ArgumentException(String.Format("Die Datei '{0}' existiert nicht", vsFilename));
         this._Filename = vsFilename;

         // Datei öffnen
         try
            {
            fs = new FileStream(vsFilename, FileMode.Open);
            sr = new StreamReader(fs);
            }
         catch
            {
            throw new Exception(string.Format("Fehler beim Öffnen der Datei '{0}'", vsFilename));
            }

         // Zeilen einlesen
         try
            {
            while(!sr.EndOfStream)
               {
               sLine = sr.ReadLine();
               //Trace.TraceInformation(this.GetType().Name + "." + MethodInfo.GetCurrentMethod().Name.ToString() + "(): "
               //   + String.Format("Less HexLine {0}: '{1}'", this.HexFileLines.Count, sLine));

               this._HexFileLines.Add(new HexFileLine(sLine));
               Thread.Sleep(0);
               }
            }
         catch(Exception e)
            {
            throw e;
            }
         finally
            {
            sr.Close();
            fs.Close();
            }
         }
      }

   // Daten einer Zeile einer Hex-Datei
   internal class HexFileLine
      {
      // Anzahl der Datenbytes
      private int _NumBytes = 0;
      public int NumBytes
         {
         get { return this._NumBytes; }
         }
      // Recordtyp der Daten
      private HexFile.eRecordType _RecordType = HexFile.eRecordType.DataRecord;
      public HexFile.eRecordType RecordType
         {
         get { return this._RecordType; }
         }
      // Zieladresse der Daten
      private System.UInt16 _Address = 0;
      public System.UInt16 Address
         {
         get { return this._Address; }
         }
      // Liste mit Daten
      private System.Byte[] _Data;
      public System.Byte[] Data
         {
         get { return this._Data; }
         }
      // Checksum
      private System.Byte _Checksum;
      public System.Byte Checksum
         {
         get { return this._Checksum; }
         }

      private const int DataOffset = 9;

      // Konstruktor
      // vsLine enthält aus Datei gelesene Textzeile
      public HexFileLine(string vsLine)
         {
         int i;

         if(vsLine.Length < 9)
            throw new Exception("HexZeile zu kurz");
         if(vsLine[0] != ':')
            throw new Exception("Zeilenbeginn != ':'");
         // Anzahl Bytes
         this._NumBytes = Convert.ToByte(vsLine.Substring(1, 2), 16);
         // Adresse
         this._Address = Convert.ToUInt16(vsLine.Substring(3, 4), 16);
         // Recordtype
         i = Convert.ToByte(vsLine.Substring(7, 2), 16);
         if(Enum.IsDefined(this._RecordType.GetType(), i))
            this._RecordType = (HexFile.eRecordType)i;
         else
            throw new Exception("Unbekannter Recordtype");

         switch(this._RecordType)
            {
            case HexFile.eRecordType.DataRecord:
               // Datenbytes einlesen
               this._Data = new System.Byte[this._NumBytes];
               for(i = 0; i < this._NumBytes; i++)
                  this._Data[i] = Convert.ToByte(vsLine.Substring(HexFileLine.DataOffset + i * 2, 2), 16);
               // Checksum einlesen
               this._Checksum = Convert.ToByte(vsLine.Substring(HexFileLine.DataOffset + this._NumBytes * 2, 2), 16);
               // TODO: Checksum prüfen
               break;
            case HexFile.eRecordType.EndOfFileRecord:
               break;
            case HexFile.eRecordType.ExtendedSegmentAddressRecord:
               this._Address = (System.UInt16)(Convert.ToUInt16(vsLine.Substring(HexFileLine.DataOffset, 4), 16) * 16);
               break;
            default:
               throw new Exception(string.Format("Recordtype '0X{0:X2}' nicht unterstützt", this._RecordType));
            }
         }
      }
   }
